/*****************************************************************************
 *                                                                           *
SC 29 Software Copyright Licencing Disclaimer:

This software module was originally developed by
  AT&T, Dolby Laboratories, Fraunhofer IIS-A

and edited by
  Yoshiaki Oikawa (Sony Corporation),
  Mitsuyuki Hatanaka (Sony Corporation)
  Coding Technologies

in the course of development of the ISO/IEC 13818-7 and ISO/IEC 14496-3 
standards for reference purposes and its performance may not have been 
optimized. This software module is an implementation of one or more tools as 
specified by the ISO/IEC 13818-7 and ISO/IEC 14496-3 standards.
ISO/IEC gives users free license to this software module or modifications 
thereof for use in products claiming conformance to audiovisual and 
image-coding related ITU Recommendations and/or ISO/IEC International 
Standards. ISO/IEC gives users the same free license to this software module or 
modifications thereof for research purposes and further ISO/IEC standardisation.
Those intending to use this software module in products are advised that its 
use may infringe existing patents. ISO/IEC have no liability for use of this 
software module or modifications thereof. Copyright is not released for 
products that do not conform to audiovisual and image-coding related ITU 
Recommendations and/or ISO/IEC International Standards.
The original developer retains full right to modify and use the code for its 
own purpose, assign or donate the code to a third party and to inhibit third 
parties from using the code for products that do not conform to audiovisual and 
image-coding related ITU Recommendations and/or ISO/IEC International Standards.
This copyright notice must be included in all copies or derivative works.
Copyright (c) ISO/IEC 1996, 2007.
 *                                                                           *
 ****************************************************************************/

#include <stdlib.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "audiofile_io.h"
#include "mps_conf.h"
#include "spatial_filereader.h"
#include "bitinput.h"
#include "adifdec.h"
#include "ct_polyphase.h"

#ifndef WIN32
#include <sys/resource.h>
#endif

#define DMX_GAIN_FIX

#define MPSCONF_VERSION "2013-01-25"

static int  parse_cmdline(int argc, char *argv[]);
static void usage(char *s);
static void print_banner(char *cmd);
static int  openAudioInput(char *umxDut, int nSmpPerFrame, AFILE **pcmStreamIn,
                           int *pnframe, int *pnchUmx, float *psampleFreq);
static void createDownMix(char *dmxPoly, AFILE *pcmStreamIn, int npart,
                          int nframe, int nSmpPerFrame, int nTimeSlots,
                          int nchDmx, int nchUmx, float sampleFreq, float dmxGain);
static int  getNchDmx(int bsTreeConfig, int *pnchDmx);
static int  getDmxGain(int bsGainDMX, float *dmxGain);
static int  getNumParts(char *fileName);
static void getTwiddle();

static int delayOffset = 0;
static int isBuriedData = 0;
static int isMpsInput = 0;

static char *pcminputfile;
static char *bitstreaminputfile;
static char *audionormal_path;
static char *audioinput_path;

#ifdef HRTF_DYNAMIC_UPDATE
HANDLE_HRTF_SOURCE hrtfSource;
#endif

int decType   = 0;

int upmixType = 0;
int hrtfModel = 0;
int binauralQuality = 0;

static Twiddle twiddle;
Twiddle *twiddleExtern = NULL;

int
main(int argc, char *argv[])
{
  AFILE *pcmStreamIn;
  char *dmxPoly = "dmx_poly.wav";
  char *umxRef  = "umx_ref.wav";
  char *bsMps   = "temp.mps";
  char *umxDut;

  int npart, nframe, nchDmx, nchUmx;
  int bsTreeConfig, bsGainDMX, bsFrameLength;
  float sampleFreq, dmxGain;
  int err;
  int max1, max2;
  double rms1, rms2;
  int nQmfBands = 64;
  int nTimeSlots;
  int nSmpPerFrame;
  int isSbr = 0;

  /* Linux: increase stack size to 32MB */
#ifndef WIN32
  struct rlimit rl = {0x2000000,-1};
  setrlimit(RLIMIT_STACK, &rl);
#endif

  parse_cmdline(argc, argv);
  umxDut = audioinput_path;

  if (upmixType == 1 || isBuriedData) { /* pcm input when blind upmix or buried data */
    MpsDecode(isBuriedData, pcminputfile, NULL, umxRef);
    npart = 1;
    nTimeSlots = 16;
    nSmpPerFrame = nTimeSlots*nQmfBands;
  }
  else {
    if (isMpsInput) {
      bsMps = bitstreaminputfile;
      GetSpatialSpecificConfigData(bsMps, &bsTreeConfig, &bsGainDMX, &bsFrameLength);
      nTimeSlots = bsFrameLength + 1;
    }
    else {
      /* Extract MPS payloads from first and second half of mp4 test file */
      err = MpsBsOpen(bsMps);
      if (err != 0) {
        PRINT(SE, "Error opening temporary bitstream file %s for writing\n", bsMps);
        exit(1);
      }
      err = ExtractMpsPayload(bitstreaminputfile);
      MpsBsClose();
      if (err) exit(1);

      /* Get data from parsing of mp4 file */
      MpsBsGetData(&bsTreeConfig, &bsGainDMX);
      nTimeSlots = 16; /* length of AAC frame */
    }
    nSmpPerFrame = nTimeSlots*nQmfBands;

    getNchDmx(bsTreeConfig, &nchDmx);
    getDmxGain(bsGainDMX, &dmxGain);
    
    if (nchDmx == 1) {
      /* Compensation for mono down mix split to 2 equal signals (L/R) when
         pass through mode of MPS decoder */
      dmxGain /= (float)sqrt(2.0);
    }

    /* Open wav output of DUT */
    err = openAudioInput(umxDut, nSmpPerFrame, &pcmStreamIn, &nframe, &nchUmx, &sampleFreq);
    if (err) {
      PRINT(SE, "Error opening %s\n", umxDut);
      exit(1); 
    }

    npart = getNumParts(bitstreaminputfile);

    /* Create downmix wav file by filtering first half (or whole if mps_qmf) of wav output
       using a polyphase correction filter */
    createDownMix(dmxPoly, pcmStreamIn, npart, nframe,  nSmpPerFrame,
                  nTimeSlots, nchDmx, nchUmx, sampleFreq, dmxGain);
    close_inputaudio (pcmStreamIn);

    if (isMpsInput) {
      MpsDecode(0, dmxPoly, bsMps, umxRef);
    }
    else if (IsSbr()) {
      isSbr = 1;
#ifndef PARTIALLY_COMPLEX
      getTwiddle();
#endif
      err = SbrMpsDecode(dmxPoly, bitstreaminputfile, umxRef);
    }
    else {  /* AAC as downmix coder */
      MpsDecode(0, dmxPoly, bsMps, umxRef);
    }
  }

  /* Compare outputs of DUT and REF decoder for first and second half
     of bitstream */
  err = DiffPcm(umxRef, umxDut, npart, nSmpPerFrame, &max1, &rms1, &max2, &rms2);

  if (err) {
    PRINT(SE, "\n\nConformance test cannot be performed\n");
  }
  else {
    if (upmixType == 1 || isBuriedData) {
      ShowConformanceResult(isBuriedData, isSbr, upmixType, binauralQuality,
                            pcminputfile, npart, max1, rms1, max2, rms2);
    }
    else {
      ShowConformanceResult(0, isSbr, upmixType, binauralQuality,
                            bitstreaminputfile, npart, max1, rms1, max2, rms2);
    }
  }

  return 0;
}

static int
parse_cmdline(int argc, char *argv[])
{
  char *cmd, *s;

  print_banner("mpsConfTool");

  cmd = argv[0];
    
  ARGBEGIN {

  default:
    PRINT(SE, "unknown option\n");
  case 's':
    s = ARGF();

    if (strcmp(s, "aac") == 0) {
      isMpsInput = 0;
      isBuriedData = 0;
    }
    else if (strcmp(s, "mps") == 0) {
      isMpsInput = 1;
      isBuriedData = 0;
    }
    else if (strcmp(s, "buried") == 0) {
      isMpsInput = 0;
      isBuriedData = 1;
    }
    else {
      PRINT(SE, "unknown bitstream source\n");
      myexit(0);
    }
    break;
  case 'h':
    usage(cmd);
    break;
  case 'd':
    s = ARGF();
    delayOffset = atoi(s);
    break;
  case 'u':
    s = ARGF();
    upmixType = atoi(s);
    break;
  case 'q':
    s = ARGF();
    binauralQuality = atoi(s);
    break;   
  }
  ARGEND

  if(argc != 2 && argc != 3) {
    usage(cmd);
  }

  if (upmixType == 1 || isBuriedData) {
    pcminputfile = argv[0];
  }
  else {
    bitstreaminputfile = argv[0];
  }

  audioinput_path = argv[1];

  if (upmixType == 2) {
    hrtfModel = 4; /* Use test hrtf */
  }

  return 0;
}

static void
usage(char *s)
{
  PRINT(SE, "Usage: %s <options> fileIn pcmIn", s);

  PRINT(SE, "\n\n");
  PRINT(SE, " fileIn                - File name of the conformance test file\n");
  PRINT(SE, " pcmIn                 - File name of the decoded input file to test (should have 24bit resolution)\n");
  PRINT(SE, "\noptions are:\n");
  PRINT(SE, " -h                    - show this help\n");
  PRINT(SE, " -s source             - bitstream source format, (default: aac)\n");
  PRINT(SE, "                         aac    - AAC or HE-AAC bitstream\n");
  PRINT(SE, "                         mps    - MPS bitstream (from mp12_mps_tool for Layer II and Layer III)\n");
  PRINT(SE, "                         buried - wav file with buried MPS bitstream \n");
  PRINT(SE, "                         wav file expected when blind upmix (-s option ignored) \n");
  PRINT(SE, " -d <delay>            - delay introduced by decoder under test, (default: 0)\n");
  PRINT(SE, " -u <upmixType>        - 0: normal, 1: blind, 2: binaural\n");
  PRINT(SE, "                         3: - 2.0 output for 5.1 bitstream\n");
  PRINT(SE, "                         4: - 5.1 output for 7.1 bitstream, (default: 0)\n");
  PRINT(SE, " -q <quality>          - Binaural quality: 0 - parametric, 1 - filtering\n");        
  PRINT(SE, "\n\nExample:\n");        
  PRINT(SE, "  %s mps_param_28.mp4 mps_param_28_toTest.wav\n", s);
  PRINT(SE, "\n");
exit(0);
}

static void
print_banner(char *cmd)
{
  PRINT(SE,"\n*********** ISO/IEC 23003-1 MPEG Surround Conformance Verification ***********\n");
  PRINT(SE,"*                                                                            *\n");
  PRINT(SE,"*                             RefSoft Version 1.19                           *\n");
#ifdef PARTIALLY_COMPLEX
  PRINT(SE,"*                                                                            *\n");
  PRINT(SE,"*                      L O W   P O W E R   D E C O D E R                     *\n");
  PRINT(SE,"*                                                                            *\n");
#endif  
  PRINT(SE,"*                    Conformance Tool Version: %s                    *\n", MPSCONF_VERSION);
  PRINT(SE,"*                                                                            *\n");
  PRINT(SE,"*                                  2013-01-25                                *\n");
  PRINT(SE,"*                                                                            *\n");
  PRINT(SE,"******************************************************************************\n");
  PRINT(SE,"\n\n");
}

static int
openAudioInput(char *umxDut, int nSmpPerFrame, AFILE **pcmStreamIn,
               int *pnframe, int *pnchUmx, float *psampleFreq) {
  audioheader audio_header;

  if (open_inputaudio (pcmStreamIn,
                       umxDut,
                       &audio_header)) {
    return 1;
  }
    
  *pnframe = audio_header.nrofsamples/nSmpPerFrame;
  *pnchUmx = audio_header.channels;
  *psampleFreq = audio_header.samplefreq;

  return 0;
}

static void
createDownMix(char *dmxPoly, AFILE *pcmStreamIn, int npart,
              int nframe, int nSmpPerFrame, int nTimeSlots,
              int nchDmx, int nchUmx, float sampleFreq, float dmxGain) {
  AFILE *pcmStreamOutDmx;
  float pcm_bufferIn[Chans*MSMP];
  float pcm_bufferOut[Chans*MSMP];
  double pcm_channel[MSMP];
#ifdef PARTIALLY_COMPLEX
  int delayQmf = 1601;
#else
  int delayQmf = 1281;
#endif
  int samples_read;
  int sample_counter;
  int delayPoly;
  int ismp, ich, iframe, ipart;

  if (binauralQuality == 1) { /* high quality binaural */
    delayQmf += 64;
  }

  pcmStreamOutDmx = AFopnWrite(dmxPoly,
                               2, /* FTW_WAVE */
                               6, /* FD_INT24 */
                               nchDmx,
                               sampleFreq,
                               0);

  /* Write polyfiltered signal of first half of DUT output twice */
  for (ipart = 0; ipart < npart; ipart++) {

    /* Get downmix */
    sample_counter = delayQmf + delayOffset;
    for (iframe = 0; iframe <= nframe/npart; iframe++) {
      samples_read = get_audiosample(pcm_bufferIn,
                                     pcmStreamIn,
                                     nchUmx,
                                     sample_counter,
                                     nSmpPerFrame);

      sample_counter += samples_read;

      if (!samples_read) {
        continue;
      }

      for (ich = 0; ich < nchDmx; ich++) {
        for (ismp = 0; ismp < nSmpPerFrame; ismp++) {
          pcm_channel[ismp] = pcm_bufferIn[ich + nchUmx*ismp];
        }

        delayPoly = QmfInverse(nTimeSlots, ich, pcm_channel);

        for (ismp = 0; ismp < nSmpPerFrame; ismp++) {
          pcm_bufferOut[ich + nchDmx*ismp] = (float)pcm_channel[ismp] / 32768.0f / dmxGain;
        }
      }

      if (iframe == 0) {
        AFfWriteData(pcmStreamOutDmx, &pcm_bufferOut[delayPoly*nchDmx],
                     (nSmpPerFrame - delayPoly)*nchDmx);
      }
      else if (iframe == nframe/npart) {
        AFfWriteData(pcmStreamOutDmx, pcm_bufferOut, delayPoly*nchDmx);
      } 
      else {
        AFfWriteData(pcmStreamOutDmx, pcm_bufferOut, nSmpPerFrame*nchDmx);
      }
    }
  }

  AFclose(pcmStreamOutDmx);
}

static int
getNchDmx(int bsTreeConfig, int *pnchDmx) {
  int nchDmxTable[] = { 1, 1, 2, 2, 2, 5, 5 };

  if (bsTreeConfig < 0 || bsTreeConfig > 6) {
    return -1;
  }

  *pnchDmx= nchDmxTable[bsTreeConfig];

  return 0;
}

static int
getDmxGain(int bsGainDMX, float *pdmxGain) {
  float dmxGainTable[] = {
    1.0f,                1.189207115002721f, 
    1.4142135623730951f, 1.681792830507429f, 
    2.0f,                2.3784142300054421f, 
    2.8284271247461903f, 4.0f
  };

  if (bsGainDMX < 0 || bsGainDMX > 7) {
    return -1;
  }

#ifdef DMX_GAIN_FIX
  if (upmixType == 2 || upmixType == 3)
    *pdmxGain = 1.0f;
  else
#endif
    *pdmxGain = dmxGainTable[bsGainDMX];

  return 0;
}

static int
getNumParts(char *fileName) {
  if (strstr(fileName, "mps_qmf")) {
    return 1;
  }
  else {
    return 2;
  }
}

static void
getTwiddle() {
  FILE *twiddlefile;

  if ((twiddlefile = fopen("twiddle.dat", "rb"))) {
    double fi, beta;        
    
    fread(&fi,   1, sizeof(double), twiddlefile);          
    fread(&beta, 1, sizeof(double), twiddlefile);          
          
    twiddle.fi   = fi;
    twiddle.beta = beta;
    twiddleExtern = &twiddle;

    PRINT(SE, "using fi = %f, beta = %f\n", fi, beta);

    fclose(twiddlefile);
  }
}




