/*****************************************************************************
 *                                                                           *
SC 29 Software Copyright Licencing Disclaimer:

This software module was originally developed by
  AT&T, Dolby Laboratories, Fraunhofer IIS-A

and edited by
  Yoshiaki Oikawa (Sony Corporation),
  Mitsuyuki Hatanaka (Sony Corporation)
  Coding Technologies

in the course of development of the ISO/IEC 13818-7 and ISO/IEC 14496-3 
standards for reference purposes and its performance may not have been 
optimized. This software module is an implementation of one or more tools as 
specified by the ISO/IEC 13818-7 and ISO/IEC 14496-3 standards.
ISO/IEC gives users free license to this software module or modifications 
thereof for use in products claiming conformance to audiovisual and 
image-coding related ITU Recommendations and/or ISO/IEC International 
Standards. ISO/IEC gives users the same free license to this software module or 
modifications thereof for research purposes and further ISO/IEC standardisation.
Those intending to use this software module in products are advised that its 
use may infringe existing patents. ISO/IEC have no liability for use of this 
software module or modifications thereof. Copyright is not released for 
products that do not conform to audiovisual and image-coding related ITU 
Recommendations and/or ISO/IEC International Standards.
The original developer retains full right to modify and use the code for its 
own purpose, assign or donate the code to a third party and to inhibit third 
parties from using the code for products that do not conform to audiovisual and 
image-coding related ITU Recommendations and/or ISO/IEC International Standards.
This copyright notice must be included in all copies or derivative works.
Copyright (c) ISO/IEC 1996, 2007.
 *                                                                           *
 ****************************************************************************/

#include "adifdec.h"
#include "bitinput.h"
#include "mps_conf.h"

char* spatialPcmOutFile;

#define MAX_INPUT_CHANNELS    ( 2)
#define MAX_TIME_SLOTS        ( 72 )

#include "libtsp.h"
#include "sac_polyphase.h"
#include "spatial_filereader.h"


/* ct_sbrdecoder.c */
void SetPcmPointers(double *pcmChannels[]);

extern HANDLE_BITINPUT static_input;
HANDLE_BYTE_READER spatialByteReader;
HANDLE_SPATIAL_DEC GetSpatialDec();

/* global, so that it can be closed in myexit() */
static HANDLE_BITINPUT input;

#define DMX_GAIN_FIX

#define MAX_QMF_BANDS 128

static void sbrMpsInit(char *pcmInFile);
static void sbrMpsTerm();
static void terminate();

static double* pcmChannels[Chans];
static float** qmfInReal[MAX_INPUT_CHANNELS];
static float** qmfInImag[MAX_INPUT_CHANNELS];
static float** inPointers[2*MAX_INPUT_CHANNELS];
static AFILE* pcmIn = 0;
static SAC_POLYPHASE_ANA_FILTERBANK *filterbank[MAX_INPUT_CHANNELS] = {0};
static float* inSamples;
static float* inSamplesDeinterleaved;
static long int nChannels;
static int qmfBands = 0;

extern int upmixType;
extern Twiddle *twiddleExtern;

int
SbrMpsDecode(char* pcmInFile, char *inputfile, char* outputfile)
{
  HANDLE_FILE_READER fileReader = NULL;
  HANDLE_BYTE_READER bitstream = NULL;
  long avgBitrate = 0;
  long maxBitrate = 0;
  int bufferFullness = -1;
  int channelConfig;
  long int offset=0;
  int samplesRead;
  int nTimeSlots = 32;
  int samplesToReadPerCall;
  int channel, ts, i, ismp, ich;
  spatialDec* ourDec;
  int secondHalf = 0;
  int channelIndex = GetChannelIndexDmx();
  spatialPcmOutFile = outputfile;
  decoder_init();

  static_input = NULL;  /* Necessary when parsing bitstream twice */
  input = bitinput_openFile(inputfile,
                            &prog_config,
                            &mc_info,
                            &channelConfig,
                            &maxBitrate,
                            &avgBitrate,
                            &bufferFullness,
                            0);
  if (input == NULL) {
    fprintf(stderr, "Error opening %s", inputfile);
    terminate();
    return(1);
  }

  decoder_setupParameters(mc_info.sampling_rate_idx);

  sbrMpsInit(pcmInFile);

  samplesToReadPerCall = nChannels * qmfBands * nTimeSlots;
  inSamplesDeinterleaved = (float*) calloc(sizeof(float), qmfBands);
  inSamples = (float*)calloc(sizeof(float), samplesToReadPerCall);

  /* main loop */
  for(bno = 0;; bno++) {
    double **output_data;
    int output_blockLen;
    int err = bitinput_getNextBlock(input);

    if (err == -2) break; /* eof */
    if (err) {
      fprintf(stderr, "Error reading %s", inputfile);
      terminate();
      return(1);
    }

    fprintf(stderr, "\rFrame [%4d]", (int)bno);
    dec_parse();

    if (IsSbrHeader() && !secondHalf) {
      offset = 0;  /* Read from beginning of file */
      secondHalf = 1;
    }

    offset += (samplesRead = AFfReadData(pcmIn, offset, inSamples, samplesToReadPerCall));
    if (samplesRead != samplesToReadPerCall) {
      fprintf(stderr, "Error reading %s", inputfile);
      terminate();
      return(1);
    }

    /* Get every second sample */
    for (ich = 0; ich < nChannels; ich++) {
      for (ismp = 0; ismp < 1024; ismp++) {
        pcmChannels[channelIndex+ich][ismp] = 32768.0 * inSamples[ich + nChannels*(2*ismp+1)];
      }
    }

    dec_decode(&output_data, &output_blockLen);
    ourDec = GetSpatialDec();

    if (!IsSbrHeader()) { /* aac core in first half of bitstream */
      SpatialDecParseFrame(ourDec);

      /* QMF analysis */
      for (channel=0; channel < nChannels; channel++) {
        /* deinterleave and scale */
        for (ts=0; ts< nTimeSlots; ts++) {
          for (i=0; i< qmfBands; i++) {
#ifdef DMX_GAIN_FIX
            if (upmixType == 2 || upmixType == 3)
              inSamplesDeinterleaved[i] = inSamples[nChannels* (ts*qmfBands+i)+channel] * 32768.0f;
            else
#endif
              inSamplesDeinterleaved[i] = inSamples[nChannels* (ts*qmfBands+i)+channel] * 32768.0f * SpatialGetClipProtectGain(ourDec);
          }

#ifdef PARTIALLY_COMPLEX
          SacCalculateAnaFilterbank( filterbank[channel],
                                     inSamplesDeinterleaved,
                                     inPointers[channel][ts],
                                     NULL );
#else
          SacCalculateAnaFilterbank( filterbank[channel],
                                     inSamplesDeinterleaved,
                                     inPointers[2*channel][ts],
                                     inPointers[2*channel+1][ts] );
#endif
        }
      }

      SpatialDecApplyFrame(ourDec,
#ifdef ALIGNMENT_FIX
                           0, /* bQmfFromSbr */
#endif
                           nTimeSlots,
                           inPointers);
    }
  }

  terminate();
  return 0;
}

static void sbrMpsInit(char *pcmInFile)
{
  long int nSamples;
  double sampleFreq;
  int i, j, channel;

  for (i=0; i < Chans; i++) {
    pcmChannels[i] = (double *)calloc(2048, sizeof(double));
  }

  SetPcmPointers(pcmChannels);

  /* allocate qmf memory */
#ifdef PARTIALLY_COMPLEX
  for (i=0 ; i < MAX_INPUT_CHANNELS ; i++) {
    qmfInReal[i] = (float **) calloc (MAX_TIME_SLOTS, sizeof (float*));
    for (j=0 ; j <  MAX_TIME_SLOTS; j++) {
      qmfInReal[i][j] = (float *) calloc (MAX_QMF_BANDS, sizeof (float));
    }
    inPointers[i] = qmfInReal[i];
    (qmfInImag);
  }
#else
  for (i=0 ; i < MAX_INPUT_CHANNELS ; i++) {
    qmfInReal[i] = (float **) calloc (MAX_TIME_SLOTS, sizeof (float*));
    qmfInImag[i] = (float **) calloc (MAX_TIME_SLOTS, sizeof (float*));
    for (j=0 ; j < MAX_TIME_SLOTS; j++) {
      qmfInReal[i][j] = (float *) calloc (MAX_QMF_BANDS, sizeof (float));
      qmfInImag[i][j] = (float *) calloc (MAX_QMF_BANDS, sizeof (float));
    }
    inPointers[2*i  ] = qmfInReal[i];
    inPointers[2*i+1] = qmfInImag[i];
  }

#endif
    pcmIn = AFopnRead(pcmInFile,
                      &nSamples,
                      &nChannels,
                      &sampleFreq,
                      0);

  qmfBands = 64;
  if (twiddleExtern != NULL) {
    SacInitAnaFilterbank(twiddleExtern, qmfBands);
  }
  else {
    SacInitAnaFilterbank(NULL, qmfBands);
  }
  for (channel = 0; channel < nChannels; channel++)
    SacOpenAnaFilterbank(&filterbank[channel]);
}

static void sbrMpsTerm() {
  int i,j, channel;

  if (inSamplesDeinterleaved != NULL) free(inSamplesDeinterleaved);
  if (inSamples != NULL) free(inSamples);

  for (i=0; i < Chans; i++) {
    if (pcmChannels[i] != NULL) free(pcmChannels[i]);
  }

#ifdef PARTIALLY_COMPLEX
  for (i=0 ; i < MAX_INPUT_CHANNELS ; i++) {
    for (j=0 ; j < MAX_TIME_SLOTS ; j++) {
      if (qmfInReal[i][j] != NULL) free(qmfInReal[i][j]);
    }
    if (qmfInReal[i] != NULL) free(qmfInReal[i]);
  }
#else
  for (i=0 ; i < MAX_INPUT_CHANNELS ; i++) {
    for (j=0 ; j < MAX_TIME_SLOTS ; j++) {
      if (qmfInReal[i][j] != NULL) free(qmfInReal[i][j]);
      if (qmfInImag[i][j] != NULL) free(qmfInImag[i][j]);
    }
    if (qmfInReal[i] != NULL) free(qmfInReal[i]);
    if (qmfInImag[i] != NULL) free(qmfInImag[i]);
  }
#endif

  for (channel = 0; channel < nChannels; channel++)
    SacCloseAnaFilterbank(filterbank[channel]);

}

static void terminate() {
  bitinput_close(input);
  decoder_free();
}

void
myexit(char *s)
{
  bitinput_close(input);
  decoder_free();
  if (s) PRINT(SE,"\n%s\n", s);
  exit(s ? 1 : 0);
}
