/**********************************************************************
 * ISO MPEG Audio Subgroup Software Simulation Group (1996)
 * ISO 13818-3 MPEG-2 Audio Decoder - Lower Sampling Frequency Extension
 *
 * $Id: common.h,v 1.1.1.1 2008-02-29 17:08:18 purnhaho Exp $
 *
 * $Log: common.h,v $
 * Revision 1.1.1.1  2008-02-29 17:08:18  purnhaho
 * initial import of mp12_mps_tool.zip from Philips 2008-02-29
 *
 * Revision 1.17  2006/03/10 15:11:55  dodenhen
 * Added MPEG_2_5 support
 *
 * Revision 1.16  2004/10/04 14:08:43  kaehleof
 * added switch for alias reduction in mixed blocks
 *
 * Revision 1.15  2004/09/29 14:57:09  kaehleof
 * moved the decoder tables into tables.c and .h;
 * no more tables/ directory needed now;
 * create_tables.c was used to create tables.c and .h
 * from the files in the tables/ directory;
 *
 * Revision 1.14  2003/11/17 15:02:15  tmm01
 *   by Ernst F. Schroeder, look for: efs
 * + added code for decoding free format,
 *   especially in Layer III.
 *   The first sync word is examined, and when
 *   free format is detected, the actual number
 *   of slots is determined from the distance
 *   of the following sync word.
 *
 * Revision 1.13  2003/11/13 12:20:30  tmm01
 *   by Ernst F. Schroeder, look for: efs
 * + added routines for CRC check in Layer III
 *   CRC errors are displayed and counted -
 *   no further action is taken
 *
 * Revision 1.12  2003/05/13 17:15:33  sps
 *  + bug fix in SubBandSynthesis() (previously introduced while changing
 *    to automatic arrays (was previously malloc))
 *
 *  + revision of the out_fifo interface (only the error code is returned,
 *    the clipping information became a function parameter)
 *
 *  + revised the code such that it can be compiled using g++
 *
 *  + functions returning an error code have now the return type
 *    ERROR_CODE. This concerns
 *     - out_fifo()
 *     - SubBandSynthesis()
 *     - III_get_side_info()
 *     - III_hufman_decode()
 *     - read_syn_window()
 *     - open_bit_stream_r()
 *     - initialize_huffman()
 *     - GetArguments()
 *     - aiff_read_headers()
 *     - aiff_seek_to_sound()
 *     - aiff_write_headers()
 *
 * Revision 1.11  2003/05/09 17:41:47  sps
 * +++ removed memory leaks.
 * +++ revised termination behaviour for unusual cases (file could not be
 * opened, no data could be written into file, memory allocation failed). A
 * couple of functions now return an error code, that leads to a clean
 * termination in the top level function.
 * +++ simplified legacy write functionality.
 * +++ removed some unnecessary "extern" declaration in function
 * declarations in header files
 *
 * Revision 1.10  2003/05/06 16:24:35  sps
 * +++ terminates if the number of channels in the bitstream exceeds the
 * number of initialized channels +++ allows to specify the number of
 * channels to be initialized using the command line option -n +++ performs
 * smart rounding from double to float prior to pass the data to the afsp
 * library write function +++ added function MyClose to assure that all
 * memory is free?d prior to exit
 *
 * Revision 1.9  2003/04/30 17:39:52  sps
 * +++ fixed aiff header for native output +++ fixed rounding for 16 bit
 * native output (by this, separated rounding in individual function
 * castDoubleToLong(), which is now called somewhat later (not anymore in
 * SubBandSynthesis(), but in out_fifo())) +++ made sure that the last
 * samples are written into the output file +++ made sure that decoding
 * starts only if file opening was successfull +++ added AFsp library
 * support, that allows the output of WAVE files. The output of raw and
 * aiff files is still supported, but is in this case also handled by the
 * AFsp library +++ Note: If the AFsp lib is used, any particular support
 * of erroneous signals is disabled +++
 *
 * Revision 1.8  2003/03/27 18:36:28  sps
 * MAX_NAME_SIZE seemed a little bit small meanwhile ...
 * -#define         MAX_NAME_SIZE           81
 * +#define         MAX_NAME_SIZE           512
 *
 * Revision 1.7  2002/11/21 18:35:25  sps
 * removed log message from header (it caused again a compiler warning)
 *
 * Revision 1.6  2002/11/21 18:31:22  sps
 *
 * Revision 1.5  2002/11/14 13:04:07  tmm01
 * warnings removed (Microsoft Visual Studio V.6.0)  23.10.2 W.Voessing
 *
 * Revision 1.3  1996/03/28 03:13:37  rowlands
 * Merged layers 1-2 and layer 3 revisions
 *
 * Revision 1.2  1996/02/14 05:18:36  rowlands
 * Cleanups.
 *
 * Revision 1.1  1996/02/14 03:45:52  rowlands
 * Initial revision
 *
 * Received from FhG
 **********************************************************************/
/**********************************************************************
 *   date   programmers         comment                               *
 * 2/25/91  Doulas Wong,        start of version 1.0 records          *
 *          Davis Pan                                                 *
 * 5/10/91  W. Joseph Carter    Reorganized & renamed all ".h" files  *
 *                              into "common.h" and "encoder.h".      *
 *                              Ported to Macintosh and Unix.         *
 *                              Added additional type definitions for *
 *                              AIFF, double/SANE and "bitstream.c".  *
 *                              Added function prototypes for more    *
 *                              rigorous type checking.               *
 * 27jun91  dpwe (Aware)        Added "alloc_*" defs & prototypes     *
 *                              Defined new struct 'frame_params'.    *
 *                              Changed info.stereo to info.mode_ext  *
 *                              #define constants for mode types      *
 *                              Prototype arguments if PROTO_ARGS     *
 * 5/28/91  Earle Jennings      added MS_DOS definition               *
 *                              MsDos function prototype declarations *
 * 7/10/91  Earle Jennings      added FLOAT definition as double      *
 *10/ 3/91  Don H. Lee          implemented CRC-16 error protection   *
 * 2/11/92  W. Joseph Carter    Ported new code to Macintosh.  Most   *
 *                              important fixes involved changing     *
 *                              16-bit ints to long or unsigned in    *
 *                              bit alloc routines for quant of 65535 *
 *                              and passing proper function args.     *
 *                              Removed "Other Joint Stereo" option   *
 *                              and made bitrate be total channel     *
 *                              bitrate, irrespective of the mode.    *
 *                              Fixed many small bugs & reorganized.  *
 *                              Modified some function prototypes.    *
 *                              Changed BUFFER_SIZE back to 4096.     *
 * 7/27/92  Michael Li          (re-)Ported to MS-DOS                 *
 * 7/27/92  Masahiro Iwadare    Ported to Convex                      *
 * 8/07/92  mc@tv.tek.com                                             *
 * 8/10/92  Amit Gulati         Ported to the AIX Platform (RS6000)   *
 *                              AIFF string constants redefined       *
 * 8/27/93 Seymour Shlien,      Fixes in Unix and MSDOS ports,        *
 *         Daniel Lauzon, and                                         *
 *         Bill Truerniet                                             *
 *--------------------------------------------------------------------*
 * 4/23/92  J. Pineda           Added code for Layer III.             *
 * 11/9/92  Amit Gulati         Added defines for layerIII stereo     *
 *                              modes.                                *
 *  8/24/93 Masahiro Iwadare    Included IS modification in Layer III.*
 *                              Changed for 1 pass decoding.          *
 *  9/07/93 Toshiyuki Ishino    Integrated Layer III with Ver 3.9.    *
 *--------------------------------------------------------------------*
 * 11/20/93 Masahiro Iwadare    Integrated Layer III with Ver 4.0.    *
 *--------------------------------------------------------------------*
 *  7/14/94 Juergen Koller      Fix for HPUX an IRIX in AIFF-Strings  *
 *--------------------------------------------------------------------*
 *  6/12/95 Soeren H. Nielsen   Bug fix in new_ext().                 *
 *  7/11/95 Soeren H. Nielsen   Added defs. for MPEG-2 LSF            *
 *--------------------------------------------------------------------*
 *          Roland Bitto	Adapted to MPEG2 low sampling rate    *
 *--------------------------------------------------------------------*
 * 01/09/06 Eugen Dodenhoeft    Added MPEG_2_5 support                *
 **********************************************************************/

#ifndef COMMON_H
#define COMMON_H

/***********************************************************************
*
*  Global Conditional Compile Switches
*
***********************************************************************/

/* #define      UNIX             Unix conditional compile switch */
/* #define      MACINTOSH        Macintosh conditional compile switch */
/* #define      MS_DOS           IBM PC conditional compile switch */
/* #define      MSC60            Compiled for MS_DOS with MSC v6.0 */
/* #define      AIX              AIX conditional compile switch    */
/* #define      CONVEX           CONVEX conditional compile switch */

#if defined(MSC60) 
  #ifndef MS_DOS
    #define MS_DOS
  #endif
  #ifndef PROTO_ARGS
    #define PROTO_ARGS
  #endif
#endif

/* 
 * Don't define FAR to far unless you're willing to clean up the 
 * prototypes
 */
#define FAR /*far*/

#ifdef __STDC__
  #ifndef PROTO_ARGS
  #define PROTO_ARGS 
  #endif
#endif

#ifdef CONVEX
  #define SEEK_SET        0
  #define SEEK_CUR        1
  #define SEEK_END        2
#endif

#ifdef _MSC_VER
  #ifndef PROTO_ARGS
  #define PROTO_ARGS 
  #endif
#endif


/***********************************************************************
*
*  Global Include Files
*
***********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "ieeefloat.h"
#include "portableio.h"

#ifdef  UNIX
#include        <unistd.h>
#endif  /* UNIX */

#ifdef __sgi
#include	<stdlib.h>
#endif

#ifdef  MACINTOSH
#include        <stdlib.h>
#include        <console.h>
#endif  /* MACINTOSH */

#ifdef  MS_DOS
#include        <stdlib.h>
#ifdef MSC60
#include        <memory.h>
#else
#include        <alloc.h>
#include        <mem.h>
#endif  /* MSC60 */
#endif  /* MS_DOS */

/***********************************************************************
*
*  Global Definitions
*
***********************************************************************/

/* General Definitions */

#ifdef  MS_DOS
#define         FLOAT                   double
#else
#define         FLOAT                   float
#endif

#define         FALSE                   0
#define         TRUE                    1
#define         NULL_CHAR               '\0'

#define         MAX_U_32_NUM            0xFFFFFFFF
#ifndef PI
#define         PI                      3.14159265358979
#endif

#define         PI4                     PI/4
#define         PI64                    PI/64
#define         LN_TO_LOG10             0.2302585093

#define         VOL_REF_NUM             0
#define         MAC_WINDOW_SIZE         24

#define         MONO                    1
#define         STEREO                  2
#define         BITS_IN_A_BYTE          8
#define         WORD                    16
#define         MAX_NAME_SIZE           512
#define         SBLIMIT                 32
#define         SSLIMIT                 18
#define         FFT_SIZE                1024
#define         HAN_SIZE                512
#define         SCALE_BLOCK             12
#define         SCALE_RANGE             64

#define         SCALE                   32768
#define         SCALE24                 32768*256  /* 24 bit output*/

#define         CRC16_POLYNOMIAL        0x8005

/* MPEG Header Definitions - Audio ID Bit Values */
#ifdef MPEG_2_5
#define         MPEG_AUDIO_ID           3
#define         MPEG_PHASE2_LSF		2	
#define		MPEG_PHASE2_5		0  	/* MPEG_2_5 extension */
#else
#define         MPEG_AUDIO_ID           1
#define         MPEG_PHASE2_LSF         0       /* 1995-07-11 SHN */
#endif

/* MPEG Header Definitions - Layer Values */ /* dodenhen 01/13/06 */
 
#ifdef MPEG_2_5
#define	        LAYER1			3
#define		LAYER2			2
#define		LAYER3			1
#else
#define	        LAYER1			0
#define		LAYER2			1
#define		LAYER3			2
#endif
 
/* MPEG Header Definitions - Mode Values */

#define         MPG_MD_STEREO           0
#define         MPG_MD_JOINT_STEREO     1
#define         MPG_MD_DUAL_CHANNEL     2
#define         MPG_MD_MONO             3

/* Mode Extention */

#define         MPG_MD_LR_LR             0
#define         MPG_MD_LR_I              1
#define         MPG_MD_MS_LR             2
#define         MPG_MD_MS_I              3

/* AIFF Definitions */

#define IFF_ID_FORM 0x464f524d /* "FORM" */
#define IFF_ID_AIFF 0x41494646 /* "AIFF" */
#define IFF_ID_COMM 0x434f4d4d /* "COMM" */
#define IFF_ID_SSND 0x53534e44 /* "SSND" */
#define IFF_ID_MPEG 0x4d504547 /* "MPEG" */

#define AIFF_FORM_HEADER_SIZE 12
#define AIFF_COMM_HEADER_SIZE 26
#define AIFF_SSND_HEADER_SIZE 16


typedef struct  blockAlign_struct {
    unsigned long   offset;
    unsigned long   blockSize;
} blockAlign;

typedef struct  IFF_AIFF_struct {
    short           numChannels;
    unsigned long   numSampleFrames;
    short           sampleSize;
    double          sampleRate;
    unsigned long   sampleType;
    blockAlign      blkAlgn;
} IFF_AIFF;

enum byte_order { order_unknown, order_bigEndian, order_littleEndian };
extern enum byte_order NativeByteOrder;

/* "bit_stream.h" Definitions */

#define         MINIMUM         4    /* Minimum size of the buffer in bytes */
#define         MAX_LENGTH      32   /* Maximum length of word written or
                                        read from bit stream */
#define         READ_MODE       0
#define         WRITE_MODE      1
#define         ALIGNING        8
#define         BINARY          0
#define         ASCII           1
#ifndef BS_FORMAT
#define         BS_FORMAT       ASCII /* BINARY or ASCII = 2x bytes */
#endif
#define         BUFFER_SIZE     4096

#define         MIN(A, B)       ((A) < (B) ? (A) : (B))
#define         MAX(A, B)       ((A) > (B) ? (A) : (B))

/***********************************************************************
*
*  Global Type Definitions
*
***********************************************************************/
typedef struct {
  char    encoded_file_name[MAX_NAME_SIZE];
  char    decoded_file_name[MAX_NAME_SIZE];
  char    mps_file_name[MAX_NAME_SIZE];
  unsigned char  need_aiff;
  unsigned char  need_esps;
  unsigned char  need_raw;
  unsigned char  need_wav;
  unsigned short topSb;
  unsigned char  Output24;		/* added to switch between 16 and 24 bit output WV 08.05.02 */
  unsigned short nrOfChannels;
  unsigned short antialias_mixed_blocks;
}Arguments_t;

/* Structure for Reading Layer II Allocation Tables from File */

typedef struct {
    unsigned int    steps;
    unsigned int    bits;
    unsigned int    group;
    unsigned int    quant;
} sb_alloc, *alloc_ptr;

typedef sb_alloc        al_table[SBLIMIT][16];

/* Header Information Structure */

typedef struct {
    int version; /* Audio version ID */
    int lay; /* Layer index */
    int error_protection; /* protection bit */
    int bitrate_index; /* bitrate index */
    int sampling_frequency; /* sampling rate index */
    int padding; /* padding bit */
    int extension; /* private bit (only informative */
    int mode; /* channel mode */
    int mode_ext; /* mode extension */
    int copyright; /* copyright bit */
    int original; /* original bit */
    int emphasis; /* emphasis */
} layer, *the_layer; 

/* Parent Structure Interpreting some Frame Parameters in Header */

typedef struct {
    layer       *header;        /* raw header information */
    int         actual_mode;    /* when writing IS, may forget if 0 chs */
    al_table    *alloc;         /* bit allocation table read in */
    int         tab_num;        /* number of table as loaded */
    int         stereo;         /* 1 for mono, 2 for stereo */
    int         stereoInit;     /* stores the initial channel configuration */
    int         jsbound;        /* first band of joint stereo coding */
    int         sblimit;        /* total number of sub bands */
} frame_params;

/* Double and SANE Floating Point Type Definitions */

typedef struct  IEEE_DBL_struct {
    unsigned long   hi;
    unsigned long   lo;
} IEEE_DBL;

typedef struct  SANE_EXT_struct {
    unsigned long   l1;
    unsigned long   l2;
    unsigned short  s1;
} SANE_EXT;


typedef enum {
  NO_ERROR = 0,
  UNKNOWN_ERROR,
  FILE_OPEN_ERROR,          /* error opening file */
  CHANNEL_CONFIG_ERROR,     /* error in channel configuration */
  SYNTHESIS_WINDOW_ERROR,   /* error in synthesis window table */
  WRITE_FILE_ERROR,         /* error writing output file */
  READ_FILE_ERROR,          /* error reading input file */
  SIDE_INFO_ERROR,          /* error in side info */
  HUFFMAN_TABLE_ERROR,      /* error in Huffman table */
  COMMAND_LINE_ERROR,       /* error in command line */
  AIFF_PROCESSING_ERROR,    /* error dealing with AIFF header */
  MEMORY_ALLOCATION_ERROR   /* error allocating memory */
} ERROR_CODE;

/* "bit_stream.h" Type Definitions */

typedef struct  bit_stream_struc {
    FILE          *pt;            /* pointer to bit stream device */
    unsigned char *buf;         /* bit stream buffer */
    int           buf_size;       /* size of buffer (in number of bytes) */
    long          totbit;         /* bit counter of bit stream */
    int           buf_byte_idx;   /* pointer to top byte in buffer */
    int           buf_bit_idx;    /* pointer to top bit of top byte in buffer */
    int           mode;           /* bit stream open in read or write mode */
    int           eob;            /* end of buffer index */
    int           eobs;           /* end of bit stream flag */
    char          format;
    
    /* format of file in rd mode (BINARY/ASCII) */
} Bit_stream_struc;

/* Layer III side information. */
typedef struct {
	unsigned part2_3_length;
	unsigned big_values;
	unsigned global_gain;
	unsigned scalefac_compress;
	unsigned window_switching_flag;
	unsigned block_type;
	unsigned mixed_block_flag;
	unsigned table_select[3];
	unsigned subblock_gain[3];
	unsigned region0_count;
	unsigned region1_count;
	unsigned preflag;
	unsigned scalefac_scale;
	unsigned count1table_select;
} my_gr_info ;

struct gr_info_s {
  unsigned part2_3_length;
  unsigned big_values;
  unsigned global_gain;
  unsigned scalefac_compress;
  unsigned window_switching_flag;
  unsigned block_type;
  unsigned mixed_block_flag;
  unsigned table_select[3];
  unsigned subblock_gain[3];
  unsigned region0_count;
  unsigned region1_count;
  unsigned preflag;
  unsigned scalefac_scale;
  unsigned count1table_select;
};

struct ch_info_s{
  unsigned scfsi[4];
  struct gr_info_s gr[2];
};

typedef struct III_side_info_s {
  unsigned main_data_begin;
  unsigned private_bits;
  struct ch_info_s ch[2];
} III_side_info_t;

/* Layer III scale factors. */

typedef struct III_scalefac_s {
  int l[23];            /* [cb] */
  int s[3][13];         /* [window][cb] */
} III_scalefac_t[2];    /* [ch] */

/***********************************************************************
*
*  Global Variable External Declarations
*
***********************************************************************/

#ifdef MPEG_2_5
extern char     *layer_names[4];
extern char     *version_names[4];
extern double   s_freq[4][4];
extern int      bitrate[4][4][15];
#else
extern char     *layer_names[3];
extern char     *version_names[2];
extern double   s_freq[2][4];
extern int      bitrate[2][3][15];
#endif

extern char     *mode_names[5];
extern double   FAR multiple[64];

/***********************************************************************
*
*  Global Function Prototype Declarations
*
***********************************************************************/

/* The following functions are in the file "common.c" */

#ifdef  PROTO_ARGS
int            js_bound(int, int);
void           hdr_to_frps(frame_params*);
void           WriteHdr(frame_params*, FILE*);
void           WriteBitAlloc(unsigned int[2][SBLIMIT], frame_params*,
                        FILE*);
void           WriteScale(unsigned int[2][SBLIMIT],
                        unsigned int[2][SBLIMIT], unsigned int[2][3][SBLIMIT],
                        frame_params*, FILE*);
void           WriteSamples(int, unsigned int FAR [SBLIMIT],
                        unsigned int[SBLIMIT], frame_params*, FILE*);
int            NumericQ(char*);
int            memcheck(char*, int, int);
void           FAR *mem_alloc(unsigned long, char*);
void           mem_free(void**);
void           double_to_extended(double*, char[10]);
void           extended_to_double(char[10], double*);
ERROR_CODE     aiff_seek_to_sound_data(FILE*);
ERROR_CODE     aiff_write_headers(FILE*, IFF_AIFF*);
void           refill_buffer(Bit_stream_struc*);
void           empty_buffer(Bit_stream_struc*, int);
ERROR_CODE     open_bit_stream_r(Bit_stream_struc*, char*, int);
void           rewind_bit_stream_r( Bit_stream_struc *bs );  /* efs 2003-11-13 */
void           close_bit_stream_r(Bit_stream_struc*);
void           alloc_buffer(Bit_stream_struc*, int);
void           desalloc_buffer(Bit_stream_struc*);
void           back_track_buffer(Bit_stream_struc*, int);
unsigned int   get1bit(Bit_stream_struc*);
void           put1bit(Bit_stream_struc*, int);
unsigned long  look_ahead(Bit_stream_struc*, int);
unsigned int   getbits_crc(Bit_stream_struc *,int N,unsigned int *);  /* efs 2003-11-11 */	
unsigned long  getbits(Bit_stream_struc*, int);
void           putbits(Bit_stream_struc*, unsigned int, int);
void           byte_ali_putbits(Bit_stream_struc*, unsigned int, int);
unsigned long  byte_ali_getbits(Bit_stream_struc*, int);
unsigned long  sstell(Bit_stream_struc*);
int            end_bs(Bit_stream_struc*);
int            seek_sync(Bit_stream_struc*, long, int);
void           I_CRC_calc(frame_params*, unsigned int[2][SBLIMIT],
                        unsigned int*);
void           II_CRC_calc(frame_params*, unsigned int[2][SBLIMIT],
                        unsigned int[2][SBLIMIT], unsigned int*);
void           update_CRC(unsigned int, unsigned int, unsigned int*);
void           read_absthr(FLOAT*, int);
unsigned int   hget1bit(); /* MI */
unsigned long  hgetbits(int);
unsigned long  hsstell();
void hputbuf(unsigned int,int);
void rewindNbytes( int );
void rewindNbits( int );

enum byte_order DetermineByteOrder();
void SwapBytesInWords( short *loc, int words );



#ifdef  MACINTOSH
void           set_mac_file_attr(char[MAX_NAME_SIZE], short, OsType,
                        OsType);
#endif
#ifdef MS_DOS
char           *new_ext(char *filename, char *extname); 
#endif

#else
int            js_bound();
void           hdr_to_frps();
void           WriteHdr();
void           WriteBitAlloc();
void           WriteScale();
void           WriteSamples();
int            NumericQ();
int            memcheck();
void           FAR *mem_alloc();
void           mem_free();
void           double_to_extended();
void           extended_to_double();
ERROR_CODE     aiff_seek_to_sound_data();
ERROR_CODE     aiff_write_headers();
void           refill_buffer();
void           empty_buffer();
ERROR_CODE     open_bit_stream_r();
void           rewind_bit_stream_r();  /* efs 2003-11-13 */
void           close_bit_stream_r();
void           alloc_buffer();
void           desalloc_buffer();
void           back_track_buffer();
unsigned int   get1bit();
void           put1bit();
unsigned long  look_ahead();
unsigned long  getbits();
void           putbits();
void           byte_ali_putbits();
unsigned long  byte_ali_getbits();
unsigned long  sstell();
int            end_bs();
int            seek_sync();
void           I_CRC_calc();
void           II_CRC_calc();
void           update_CRC();
void           read_absthr();

unsigned int   hget1bit();
unsigned long  hgetbits();
unsigned long  hsstell();
void           hputbuf();
void           rewindNbytes();

#ifdef MS_DOS
char           *new_ext(); 
#endif
#endif
#endif

