/*******************************************************************************
This software module was originally developed by

Agere Systems, Coding Technologies, Fraunhofer IIS, Philips

and edited by

-

in the course of development of ISO/IEC 23003 for reference purposes and its
performance may not have been optimized. This software module is an
implementation of one or more tools as specified by ISO/IEC 23003. ISO/IEC gives
You a royalty-free, worldwide, non-exclusive, copyright license to copy,
distribute, and make derivative works of this software module or modifications
thereof for use in implementations of ISO/IEC 23003 in products that satisfy
conformance criteria (if any). Those intending to use this software module in
products are advised that its use may infringe existing patents. ISO/IEC have no
liability for use of this software module or modifications thereof. Copyright is
not released for products that do not conform to audiovisual and image-coding
related ITU Recommendations and/or ISO/IEC International Standards.

Agere Systems, Coding Technologies, Fraunhofer IIS, Philips retain full right to
modify and use the code for its (their) own purpose, assign or donate the code
to a third party and to inhibit third parties from using the code for products
that do not conform to MPEG-related ITU Recommendations and/or ISO/IEC
International Standards. This copyright notice must be included in all copies or
derivative works.

Copyright (c) ISO/IEC 2009.
*******************************************************************************/




#include "sac_dec.h"
#include "sac_process.h"
#include "sac_bitdec.h"
#include "sac_decor.h"
#include "sac_hybfilter.h"
#include "sac_types.h"
#include "sac_smoothing.h"
#include "sac_tonality.h"
#include "sac_TPprocess.h"
#include "sac_reshapeBBEnv.h"
#include "sac_nlc_dec.h"
#include "sac_mdct2qmf.h"

#include "sac_blind.h"

#include "sac_calcM1andM2.h"
#include "sac_hrtf.h"
#include "sac_parallelReverb.h"

#include "math.h"

#include <assert.h>

#define DECORR_FIX 





#ifdef PARTIALLY_COMPLEX
#ifdef ALIGNMENT_FIX
void Cos2Exp ( const float prData[][PC_NUM_BANDS], const float  prDataAhead[][PC_NUM_BANDS], float *piData, int nDim );
#else
void Cos2Exp ( const float prData[][PC_NUM_BANDS], float *piData, int nDim );
#endif
void QmfAlias( spatialDec  *self );
#endif

extern Twiddle *twiddleExtern;


spatialDec* SpatialDecOpen(HANDLE_BYTE_READER bitstream,
                           char* spatialPcmOutFileName,
                           int spatialPcmOutFileResolution,
                           int samplingFreq,
                           int nInChannels,
                           int* pTimeSlots,
                           int nQmfBands,
                           int nDecType,
                           int nUpmixType,
                           int nHrtfModel,
#ifdef HRTF_DYNAMIC_UPDATE
                           HANDLE_HRTF_SOURCE hrtfSource,
#endif 
                           int nBinauralQuality) {


  int nCh, i, j, k, sacHeaderLen;
  int pcmResolution = 5;  

  spatialDec* self =0;
  self = (spatialDec*) calloc (sizeof(spatialDec),1);

  if (self==0) return 0;
  self->bitstream =0;
  self->outputFile =0;

  
  self->numParameterSets = 1;

  self->decType = nDecType;

  self->upmixType = nUpmixType;

  self->qmfBands = nQmfBands;

#ifdef HRTF_DYNAMIC_UPDATE
  self->hrtfSource = hrtfSource;
#endif 

  self->binauralQuality = nBinauralQuality;


  

  if (self->upmixType != 1) {
    self->bitstream = s_bitinput_open(bitstream);
    if (self->bitstream ==0) {
        SpatialDecClose(self);
        return 0;
    }
  }

  

  if (self->upmixType != 1) {
    self->sacTimeAlignFlag = s_GetBits(self->bitstream, 1);

    sacHeaderLen = s_GetBits(self->bitstream, 7);
    if (sacHeaderLen == 127) {
      sacHeaderLen += s_GetBits(self->bitstream, 16);
    }

    SpatialDecParseSpecificConfig(self, sacHeaderLen, 0);

    s_byteAlign(self->bitstream);

    if (self->sacTimeAlignFlag) {
      self->sacTimeAlign = s_GetBits(self->bitstream, 16);
      if (self->sacTimeAlign != 0) {
        myexit("ERROR: sacTimeAlign != 0 not implemented");
      }
    }
  }
  else {
    SpatialDecDefaultSpecificConfig(self, samplingFreq);
  }

  
  for(i=0; i<MAX_OUTPUT_CHANNELS_AT; i++)
    for(j=0; j<MAX_PARAMETER_BANDS; j++)
      self->prevGainAT[i][j] = 1.0f;

  SpatialDecDecodeHeader(self);


  assert(samplingFreq == self->samplingFreq);
  assert(nInChannels == self->numInputChannels);
  assert((nQmfBands == 32)||(nQmfBands == 64)||(nQmfBands == 128));


  *pTimeSlots = self->timeSlots;

  

  SpatialDecInitResidual(self);
  SPACE_MDCT2QMF_Create(self);

  
  initSpatialTonality(self);
  initParameterSmoothing(self);

  

  SacInitBBEnv(self);

  if (twiddleExtern != NULL) {
    SacInitSynFilterbank(twiddleExtern, self->qmfBands);
  }
  else {
    SacInitSynFilterbank(NULL, self->qmfBands);
  }

  for (nCh = 0; nCh < self->numOutputChannelsAT; nCh++) {
    SacOpenSynFilterbank(&self->qmfFilterState[nCh]);
  }

  
  for (nCh = 0; nCh < self->numInputChannels; nCh++) {
    SacInitAnaHybFilterbank(&self->hybFilterState[nCh]);
  }

  if (self->residualCoding) {
    int offset = self->numInputChannels;
    for (nCh = 0; nCh < self->numOttBoxes + self->numTttBoxes; nCh++) {
      if (self->resBands[nCh] > 0) {
        SacInitAnaHybFilterbank(&self->hybFilterState[offset + nCh]);
      }
    }
  }

  if (self->arbitraryDownmix == 2) {
    int offset = self->numInputChannels + self->numOttBoxes + self->numTttBoxes;
    for (nCh = 0; nCh < self->numInputChannels; nCh++) {
      SacInitAnaHybFilterbank(&self->hybFilterState[offset + nCh]);
    }
  }

  for (nCh=0; nCh<self->numOutputChannelsAT; nCh++)
  {
    SacInitSynHybFilterbank();
  }

#ifdef PARTIALLY_COMPLEX
  {
    int seedVec[MAX_NO_DECORR_CHANNELS];
    int psIdx = 0;

    for (k=0; k<MAX_NO_DECORR_CHANNELS; k++){
      seedVec[k] = k;
    }


    if (self->treeConfig == TREE_5151){
      seedVec[0] = 0;
      seedVec[1] = 1;
      seedVec[2] = 2;
      seedVec[3] = 2;
      psIdx = 2;
#ifdef DECORR_FIX
      if (self->upmixType == 2) { 
        psIdx = 0;
      }
#endif
    }
    if (self->treeConfig == TREE_5152){
      seedVec[0] = 0;
      seedVec[1] = 2;
      seedVec[2] = 1;
      seedVec[3] = 1;
      psIdx = 2;
#ifdef DECORR_FIX
      if (self->upmixType == 2) {
        psIdx = 0;
      }
#endif
    }
    if (self->treeConfig == TREE_525){
      seedVec[0] = 0;
      seedVec[1] = 0;
      psIdx = 0;
    }
    if (self->treeConfig == TREE_7271){
      seedVec[0] = 1;
      seedVec[1] = 1;
      seedVec[2] = 2;
      seedVec[3] = 0;
      seedVec[4] = 0;
      psIdx = 1;
    }
    if (self->treeConfig == TREE_7272){
      seedVec[0] = 1;
      seedVec[1] = 1;
      seedVec[2] = 2;
      seedVec[3] = 0;
      seedVec[4] = 0;
      psIdx = 1;
    }
    if (self->treeConfig == TREE_7571){
      seedVec[0] = 0;
      seedVec[1] = 0;
      psIdx = 0;
    }
    if (self->treeConfig == TREE_7572){
      seedVec[0] = 0;
      seedVec[1] = 0;
      psIdx = 0;
    }

    for (k=0; k<MAX_NO_DECORR_CHANNELS; k++)
    {
      int errCode;

      errCode = SpatialDecDecorrelateCreate(  &(self->apDecor[k]),
                                                  self->hybridBands,
                                                  seedVec[k],
                                                  self->decType,
                                                  seedVec[k] == psIdx,
                                                  self->decorrConfig
                                               );
      if (errCode != ERR_NONE) break;
    }
  }

#else

  for (k=0; k<MAX_NO_DECORR_CHANNELS; k++)
  {
    int errCode, idec;

#ifdef DECORR_FIX
    if (self->upmixType == 3) {
      idec = 0;
    }
    else {
      idec = k;
    }
#else
    idec = k;
#endif

    errCode = SpatialDecDecorrelateCreate(  &(self->apDecor[k]),
                                                self->hybridBands,
                                                idec,
                                                self->decType,
                                                0,
                                                self->decorrConfig
                                             );
    if (errCode != ERR_NONE) break;
  }
#endif

  if (self->upmixType == 1) {
    SpatialDecInitBlind(self);
  }

  

  
  if ( 24 == spatialPcmOutFileResolution )
    pcmResolution = 6;

  

   self->outputFile = AFopnWrite(spatialPcmOutFileName,
                                2, 
                                pcmResolution, 
                                self->numOutputChannelsAT,
                                (double)self->samplingFreq,
                                0);



  SpatialDecInitM1andM2(self, nHrtfModel);


  
  BinauralInitParallelReverb(self);


  
  self->parseNextBitstreamFrame=1;


  
  for(k = 0; k < 2; k++) {

    SPATIAL_BS_FRAME bsFrame = self->bsFrame[k];

    for(i = 0; i < MAX_NUM_OTT; i++) {
      for(j = 0; j < MAX_PARAMETER_BANDS; j++) {
        bsFrame.ottCLDidxPrev       [i][j] = 0;
        bsFrame.ottICCidxPrev       [i][j] = 0;
        bsFrame.cmpOttCLDidxPrev    [i][j] = 0;
        bsFrame.cmpOttICCidxPrev    [i][j] = 0;
#ifdef VERIFICATION_TEST_COMPATIBILITY
        bsFrame.cmpOttICCtempidxPrev[i][j] = 0;
#endif
      }
    }
    for(i = 0; i < MAX_NUM_TTT; i++) {
      for(j = 0; j < MAX_PARAMETER_BANDS; j++) {
        bsFrame.tttCPC1idxPrev   [i][j] = 0;
        bsFrame.tttCPC2idxPrev   [i][j] = 0;
        bsFrame.tttCLD1idxPrev   [i][j] = 0;
        bsFrame.tttCLD2idxPrev   [i][j] = 0;
        bsFrame.tttICCidxPrev    [i][j] = 0;
        bsFrame.cmpTttCPC1idxPrev[i][j] = 0;
        bsFrame.cmpTttCPC2idxPrev[i][j] = 0;
        bsFrame.cmpTttCLD1idxPrev[i][j] = 0;
        bsFrame.cmpTttCLD2idxPrev[i][j] = 0;
        bsFrame.cmpTttICCidxPrev [i][j] = 0;
      }
    }
    for(i = 0; i < MAX_INPUT_CHANNELS; i++) {
      for(j = 0; j < MAX_PARAMETER_BANDS; j++) {
        bsFrame.arbdmxGainIdxPrev[i][j]    = 0;
        bsFrame.cmpArbdmxGainIdxPrev[i][j] = 0;
      }
    }

  }

  return self;
}



void SpatialDecClose(spatialDec* self) {
  int k;

  if (self) {
    for (k = 0; k < self->numOutputChannelsAT; k++) {
      SacCloseSynFilterbank(self->qmfFilterState[k]);
    }

    SacCloseAnaHybFilterbank(); 
    SacCloseSynHybFilterbank();

    for (k = 0; k < MAX_NO_DECORR_CHANNELS; k++) {
      SpatialDecDecorrelateDestroy(&self->apDecor[k]);
    }

    if(self->bsFrame[0].resData.residualDataIntern)
    {
      free(self->bsFrame[0].resData.residualDataIntern);
      self->bsFrame[0].resData.residualDataIntern=0;
    }

    if (self->upmixType == 1) {
      SpatialDecCloseBlind(self);
    }

    if (self->upmixType == 2) {
      SpatialDecCloseHrtfFilterbank(self->hrtfFilter);

#ifdef HRTF_DYNAMIC_UPDATE
      if (self->hrtfSource != NULL) {
        free((HRTF_DATA*) self->hrtfData);
      }
#endif 
    }

	BinauralDestroyParallelReverb(self);

    if (self->outputFile) AFclose(self->outputFile);
    if (self->bitstream) s_bitinput_close(self->bitstream);
    free(self);
  }
  return;
}





void SpatialDecApplyFrame(spatialDec* self,
#ifdef ALIGNMENT_FIX
                          int bQmfFromSbr,
#endif
                          int inTimeSlots,
                          float** pointers[4]   
                          )
{

  float pcmOutBuf[MAX_OUTPUT_CHANNELS*MAX_TIME_SLOTS*MAX_NUM_QMF_BANDS];
  int ch,ts,sam,qs,n;

#ifdef ALIGNMENT_FIX
  self->bQmfFromSbr = bQmfFromSbr;
#endif

  

  assert(self->curTimeSlot+inTimeSlots <= self->timeSlots);


#ifdef PARTIALLY_COMPLEX
  
  for (ts = 0; ts < inTimeSlots; ts++) {
    const float invSqrt2 = (float) (1.0 / sqrt(2.0));

    for (ch = 0; ch < self->numInputChannels; ch++) {
      for (qs = 0; qs < PC_NUM_BANDS; qs++) {
        for (n = 0; n < (PC_FILTERLENGTH - 1); n++) {
          self->qmfCos2ExpInput[ch][n][qs] = self->qmfCos2ExpInput[ch][n + 1][qs];
        }
#ifdef ALIGNMENT_FIX
        for (n = 0; n < (PC_FILTERLENGTH - 1); n++) {
          self->qmfCos2ExpInputAhead[ch][n][qs] = self->qmfCos2ExpInputAhead[ch][n + 1][qs];
        }

        self->qmfCos2ExpInput[ch][PC_FILTERLENGTH - 1][qs] = pointers[ch][ts][qs];
        if (self->bQmfFromSbr) {
          self->qmfCos2ExpInputAhead[ch][PC_FILTERLENGTH - 1][qs] = pointers[ch][ts+HYBRID_FILTER_DELAY][qs];
        }
        else {
          self->qmfCos2ExpInputAhead[ch][PC_FILTERLENGTH - 1][qs] = pointers[ch][ts][qs];
        }
#else
        self->qmfCos2ExpInput[ch][PC_FILTERLENGTH - 1][qs] = pointers[ch][ts][qs];
#endif
      }

#ifdef ALIGNMENT_FIX
      Cos2Exp(self->qmfCos2ExpInput[ch], self->qmfCos2ExpInputAhead[ch], 
              self->qmfInputImag[ch][self->curTimeSlot+ts], PC_NUM_BANDS);
#else
      Cos2Exp(self->qmfCos2ExpInput[ch], self->qmfInputImag[ch][self->curTimeSlot+ts], PC_NUM_BANDS);
#endif

      for (qs = 0; qs < PC_NUM_BANDS; qs++) {
        self->qmfInputReal[ch][self->curTimeSlot+ts][qs]  = invSqrt2 * self->qmfInputDelayReal[ch][self->qmfInputDelayIndex][qs];
        self->qmfInputImag[ch][self->curTimeSlot+ts][qs] *= invSqrt2;

#ifdef ALIGNMENT_FIX
        if (self->bQmfFromSbr && qs < QMF_BANDS_TO_HYBRID) {
          self->qmfInputDelayReal[ch][self->qmfInputDelayIndex][qs] = pointers[ch][ts+HYBRID_FILTER_DELAY][qs];
        }
        else {
          self->qmfInputDelayReal[ch][self->qmfInputDelayIndex][qs] = pointers[ch][ts][qs];
        }
#else
        self->qmfInputDelayReal[ch][self->qmfInputDelayIndex][qs] = pointers[ch][ts][qs];
#endif
      }

      for (qs = PC_NUM_BANDS; qs < self->qmfBands; qs++) {
        self->qmfInputReal[ch][self->curTimeSlot+ts][qs] = self->qmfInputDelayReal[ch][self->qmfInputDelayIndex][qs];
        self->qmfInputImag[ch][self->curTimeSlot+ts][qs] = 0;

        self->qmfInputDelayReal[ch][self->qmfInputDelayIndex][qs] = pointers[ch][ts][qs];
      }
    }

    self->qmfInputDelayIndex++;

    if (self->qmfInputDelayIndex == PC_FILTERDELAY) {
      self->qmfInputDelayIndex = 0;
    }
  }
#else
  
  if (self->upmixType == 1) {
    for (ts = 0; ts < inTimeSlots; ts++) {
      for (ch = 0; ch < self->numInputChannels; ch++) {
        for (qs = 0; qs < self->qmfBands; qs++) {
#ifdef ALIGNMENT_FIX
          if (self->bQmfFromSbr && qs < QMF_BANDS_TO_HYBRID) {
            self->qmfInputReal[ch][self->curTimeSlot+ts][qs] = pointers[2*ch][ts+HYBRID_FILTER_DELAY][qs];
            self->qmfInputImag[ch][self->curTimeSlot+ts][qs] = pointers[2*ch+1][ts+HYBRID_FILTER_DELAY][qs];
          }
          else{
            self->qmfInputReal[ch][self->curTimeSlot+ts][qs] = pointers[2*ch][ts][qs];
            self->qmfInputImag[ch][self->curTimeSlot+ts][qs] = pointers[2*ch+1][ts][qs];
          }
#else
          self->qmfInputReal[ch][self->curTimeSlot+ts][qs] = pointers[2*ch][ts][qs];
          self->qmfInputImag[ch][self->curTimeSlot+ts][qs] = pointers[2*ch+1][ts][qs];
#endif
        }
      }
    }
  }
  else {
    for (ts = 0; ts < inTimeSlots; ts++) {
      for (ch = 0; ch < self->numInputChannels; ch++) {
        for (qs = 0; qs < self->qmfBands; qs++) {
          self->qmfInputReal[ch][self->curTimeSlot+ts][qs] = self->qmfInputDelayReal[ch][self->qmfInputDelayIndex][qs];
          self->qmfInputImag[ch][self->curTimeSlot+ts][qs] = self->qmfInputDelayImag[ch][self->qmfInputDelayIndex][qs];

#ifdef ALIGNMENT_FIX
          if (self->bQmfFromSbr && qs < QMF_BANDS_TO_HYBRID) {
            self->qmfInputDelayReal[ch][self->qmfInputDelayIndex][qs] = pointers[2*ch][ts+HYBRID_FILTER_DELAY][qs];
            self->qmfInputDelayImag[ch][self->qmfInputDelayIndex][qs] = pointers[2*ch+1][ts+HYBRID_FILTER_DELAY][qs];
          }
          else{
            self->qmfInputDelayReal[ch][self->qmfInputDelayIndex][qs] = pointers[2*ch][ts][qs];
            self->qmfInputDelayImag[ch][self->qmfInputDelayIndex][qs] = pointers[2*ch+1][ts][qs];
          }
#else
          self->qmfInputDelayReal[ch][self->qmfInputDelayIndex][qs] = pointers[2*ch][ts][qs];
          self->qmfInputDelayImag[ch][self->qmfInputDelayIndex][qs] = pointers[2*ch+1][ts][qs];
#endif
        }
      }

      self->qmfInputDelayIndex++;

      if (self->qmfInputDelayIndex == PC_FILTERDELAY) {
        self->qmfInputDelayIndex = 0;
      }
    }
  }
#endif

  self->curTimeSlot += inTimeSlots;

  
  if (self->curTimeSlot < self->timeSlots)
    return;

  self->curTimeSlot = 0;

  SpatialDecDecodeFrame(self);

  SpatialDecMDCT2QMF(self);

  SpatialDecHybridQMFAnalysis(self);

  if (self->upmixType == 1) {
    SpatialDecApplyBlind(self);
  }

#ifdef PARTIALLY_COMPLEX
  QmfAlias(self);
#endif

  SpatialDecCreateX(self);

  SpatialDecCalculateM1andM2(self);
  SpatialDecSmoothM1andM2(self);


  if(1==self->decodeAfterConfigHasChangedFlag)
  {
    int i,j,k;
    
    self->decodeAfterConfigHasChangedFlag=0;

    for(i=0; i< MAX_PARAMETER_BANDS; i++){
      for(j = 0; j < MAX_M_OUTPUT; j++){
        for(k=0; k < MAX_M_INPUT; k++){
          self->M0arbdmxPrev[i][j][k]    = self->M0arbdmx[0][i][j][k];
          self->M1paramRealPrev[i][j][k] = self->M1paramReal[0][i][j][k];
          self->M1paramImagPrev[i][j][k] = self->M1paramImag[0][i][j][k];
          
          self->M2decorRealPrev[i][j][k] = self->M2decorReal[0][i][j][k];
          self->M2decorImagPrev[i][j][k] = self->M2decorImag[0][i][j][k];
          
          self->M2residRealPrev[i][j][k] = self->M2residReal[0][i][j][k];
          self->M2residImagPrev[i][j][k] = self->M2residImag[0][i][j][k];
        }
      }
    }
  }
  

  SpatialDecApplyM1(self);

  
  SpatialDecCreateW(self);

  SpatialDecApplyM2(self);



  
  if ( (self->upmixType != 2) && (self->upmixType != 3) ) {
    
    if (self->tempShapeConfig == 2) {
      spatialDecReshapeBBEnv(self);
    }
  }

  tpProcess(self);


  if(!self->bsConfig.arbitraryTree && self->upmixType != 2 && self->upmixType != 3) {
    
    for (n=0; n< self->frameLength; n++) {
      self->timeOut[3][n] *= self->lfeGain;       
      self->timeOut[4][n] *= self->surroundGain;  
      self->timeOut[5][n] *= self->surroundGain;  
    
      if(self->treeConfig == 4 || self->treeConfig == 6){
        self->timeOut[6][n] *= self->surroundGain;  
        self->timeOut[7][n] *= self->surroundGain;  
      }
    }
  }

  BinauralApplyParallelReverb(self);

  

  for (ch=0; ch< self->numOutputChannelsAT; ch++) {
    for (sam=0; sam<self->frameLength; sam++) {
      pcmOutBuf[sam*self->numOutputChannelsAT + ch] =
        (float)(self->timeOut[ch][sam] / 32768.0f);
    }
  }

  SpatialDecUpdateBuffers(self);

  

  AFfWriteData(self->outputFile, pcmOutBuf,self->frameLength*self->numOutputChannelsAT);

  self->parseNextBitstreamFrame=1;
}



float SpatialGetClipProtectGain(spatialDec* self) {
  return self->clipProtectGain;
}


void SpatialDecPartiallyReinit(spatialDec *self)
{
  
  int i,j,k;
  int qb;

  self->decodeAfterConfigHasChangedFlag=1;

  if(self->bsFrame[0].resData.residualDataIntern)
  {
    free(self->bsFrame[0].resData.residualDataIntern);
    self->bsFrame[0].resData.residualDataIntern=0;
  }

  SpatialDecInitResidual(self);
  
  
  for(qb = 0; qb < self->hybridBands; qb++){
    int parBandindx = self->kernels[qb][0];
    int qmfindx=SacGetQmfSubband( qb );
    
    for ( i = 0; i < MAX_RESIDUAL_CHANNELS; i++)
    {
      if((parBandindx>=self->bsConfig.bsResidualBandsPrevious[i]) && (parBandindx<=self->bsConfig.bsResidualBands[i]))
      {
        for ( j = 0; j < 2 * MAX_TIME_SLOTS; j++ )
        {
          self->qmfResidualReal[i][j][qmfindx]=0.0f; 
          self->qmfResidualImag[i][j][qmfindx]=0.0f; 
        }
      }
    }
  }

  for(i=0;i<MAX_NUM_OTT;i++) {
    for(j=0;j<MAX_PARAMETER_SETS;j++) {
      for(k=0;k<MAX_PARAMETER_BANDS;k++) {
        self->bsFrame->ottICCdiffidx[i][j][k]=0.f;
      }
    }
  }
 
  
  initSpatialTonality(self);
  initParameterSmoothing(self);

  

  SacInitBBEnv(self);

  

  for (k = 0; k < MAX_NO_DECORR_CHANNELS; k++) {
    SpatialDecDecorrelateDestroy(&self->apDecor[k]);
  }

#ifdef PARTIALLY_COMPLEX
  {
    int seedVec[MAX_NO_DECORR_CHANNELS];
    int psIdx = 0;

    for (k=0; k<MAX_NO_DECORR_CHANNELS; k++){
      seedVec[k] = k;
    }

    if (self->treeConfig == TREE_5151){
      seedVec[0] = 0;
      seedVec[1] = 1;
      seedVec[2] = 2;
      seedVec[3] = 2;
      psIdx = 2;
#ifdef DECORR_FIX
      if (self->upmixType == 2) { 
        psIdx = 0;
      }
#endif
    }
    if (self->treeConfig == TREE_5152){
      seedVec[0] = 0;
      seedVec[1] = 2;
      seedVec[2] = 1;
      seedVec[3] = 1;
      psIdx = 2;
#ifdef DECORR_FIX
      if (self->upmixType == 2) {
        psIdx = 0;
      }
#endif
    }
    if (self->treeConfig == TREE_525){
      seedVec[0] = 0;
      seedVec[1] = 0;
      psIdx = 0;
    }
    if (self->treeConfig == TREE_7271){
      seedVec[0] = 1;
      seedVec[1] = 1;
      seedVec[2] = 2;
      seedVec[3] = 0;
      seedVec[4] = 0;
      psIdx = 1;
    }
    if (self->treeConfig == TREE_7272){
      seedVec[0] = 1;
      seedVec[1] = 1;
      seedVec[2] = 2;
      seedVec[3] = 0;
      seedVec[4] = 0;
      psIdx = 1;
    }
    if (self->treeConfig == TREE_7571){
      seedVec[0] = 0;
      seedVec[1] = 0;
      psIdx = 0;
    }
    if (self->treeConfig == TREE_7572){
      seedVec[0] = 0;
      seedVec[1] = 0;
      psIdx = 0;
    }

    for (k=0; k<MAX_NO_DECORR_CHANNELS; k++)
    {
      int errCode;

      errCode = SpatialDecDecorrelateCreate(  &(self->apDecor[k]),
                                                  self->hybridBands,
                                                  seedVec[k],
                                                  self->decType,
                                                  seedVec[k] == psIdx,
                                                  self->decorrConfig
                                               );
      if (errCode != ERR_NONE) break;
    }
  }

#else

  for (k=0; k<MAX_NO_DECORR_CHANNELS; k++)
  {
    int errCode, idec;

#ifdef DECORR_FIX
    if (self->upmixType == 3) {
      idec = 0;
    }
    else {
      idec = k;
    }
#else
    idec = k;
#endif

    errCode = SpatialDecDecorrelateCreate(  &(self->apDecor[k]),
                                                self->hybridBands,
                                                idec,
                                                self->decType,
                                                0,
                                                self->decorrConfig
                                             );
  }
#endif

 
  for(i=0; i< MAX_PARAMETER_BANDS; i++){
    for(j = 0; j < MAX_M_OUTPUT; j++){
      for(k=0; k < MAX_M_INPUT; k++){
        self->M0arbdmxPrev[i][j][k]    = self->M0arbdmx[0][i][j][k];
        self->M1paramRealPrev[i][j][k] = self->M1paramReal[0][i][j][k];
        self->M1paramImagPrev[i][j][k] = self->M1paramImag[0][i][j][k];
        
        self->M2decorRealPrev[i][j][k] = self->M2decorReal[0][i][j][k];
        self->M2decorImagPrev[i][j][k] = self->M2decorImag[0][i][j][k];
        
        self->M2residRealPrev[i][j][k] = self->M2residReal[0][i][j][k];
        self->M2residImagPrev[i][j][k] = self->M2residImag[0][i][j][k];
      }
    }
  }
}
