/*******************************************************************************
This software module was originally developed by

Agere Systems, Coding Technologies, Fraunhofer IIS, Philips

and edited by

-

in the course of development of ISO/IEC 23003 for reference purposes and its
performance may not have been optimized. This software module is an
implementation of one or more tools as specified by ISO/IEC 23003. ISO/IEC gives
You a royalty-free, worldwide, non-exclusive, copyright license to copy,
distribute, and make derivative works of this software module or modifications
thereof for use in implementations of ISO/IEC 23003 in products that satisfy
conformance criteria (if any). Those intending to use this software module in
products are advised that its use may infringe existing patents. ISO/IEC have no
liability for use of this software module or modifications thereof. Copyright is
not released for products that do not conform to audiovisual and image-coding
related ITU Recommendations and/or ISO/IEC International Standards.

Agere Systems, Coding Technologies, Fraunhofer IIS, Philips retain full right to
modify and use the code for its (their) own purpose, assign or donate the code
to a third party and to inhibit third parties from using the code for products
that do not conform to MPEG-related ITU Recommendations and/or ISO/IEC
International Standards. This copyright notice must be included in all copies or
derivative works.

Copyright (c) ISO/IEC 2009.
*******************************************************************************/

#include "sac_resdecode.h"
#include "sac_reshuffdec.h"
#include "sac_reshuffdec_intern.h"
#include <assert.h>

extern long frameCounter;

static int sampling_boundarys[16] = {
  92017, 75132, 55426, 46009, 37566, 27713, 23004, 18783,
  13856, 11502, 9391, 0, 0, 0, 0, 0
};




static void
synt(Info *info, byte *group, byte *mask, Float *right, Float *left)
{
  Float vrr, vrl, *r, *l;
  short *band;
  int i, n, nn, b, bb, nband;

  
  bb = 0;
  for(b = 0; b < info->nsbk; ){
    nband = info->sfb_per_sbk[b];
    band = info->sbk_sfb_top[b];

    b = *group++;		
    for(; bb < b; bb++){	
      n = 0;
      for(i = 0; i < nband; i++){
        nn = band[i];	
        if (mask[i]){
          r = right + n;
          l = left + n;
          for(; n < nn; n++){	
            vrr = *r;
            vrl = *l;
            *l = vrr + vrl;
            *r = vrl - vrr;
            r++;
            l++;
          }
        }
        n = nn;
      }
      right += info->bins_per_sbk[bb];
      left += info->bins_per_sbk[bb];

    }
    mask += info->sfb_per_sbk[bb-1];
  }
  
}

static void applyTools(Float* coef,
                       TNS_frame_info* tns,
                       byte max_sfb,
                       Info* info){

  int i, j, ch=0;

  
  for (i=j=0; i<tns->n_subblocks; i++) {
    if (debug['T']) {
      PRINT(SE, "%ld %d %d\n", frameCounter, ch, i);
      s_print_tns( &(tns->info[i]));
    }

    s_tns_decode_subblock(&coef[j],
                        max_sfb,
                        info->sbk_sfb_top[i],
                        info->islong,
                        &(tns->info[i]) );

    j += info->bins_per_sbk[i];
  }

}

BLOCKTYPE decodeIcs(spatialDec* self,
                    int channel,
                    int channelFrame,
                    int extra) {

  int i;
  int common_window = 0;
  RESIDUAL_DATA_INTERN* rdi =
    (RESIDUAL_DATA_INTERN*) self->bsFrame[0].resData.residualDataIntern;

  RDI_CHANNEL* rdic = &(rdi->rdiChannel[channel]);

  s_huffdecode(ID_SCE,
               common_window,
               &(rdic->win),
               &(rdic->wndShape),
               rdic->cb_map,
               rdic->factors,
               rdic->group,
               &(rdic->hasmask),
               rdic->mask,
               &(rdic->max_sfb),
               &(rdic->tns),
               rdic->coef,
               &(rdic->info),
               &(rdi->winmap[0])
               );


  applyTools(rdic->coef,
             &rdic->tns,
             rdic->max_sfb,
             &rdic->info);

  
  self->resMdctPresent[channel][channelFrame] = 1;

  if (extra==1) 
	  assert(self->resBlocktype[channel][channelFrame] == rdic->win);
  else
	  self->resBlocktype[channel][channelFrame] = rdic->win;

  for (i=0; i< BLOCK_LEN_LONG; i++) {
    self->resMDCT[channel][channelFrame][extra*MAX_MDCT_COEFFS + i]=
      (float)rdic->coef[i]; 
  }


  return rdic->win;
}

BLOCKTYPE decodeCpe(spatialDec* self,
                    int channel,
                    int channelFrame,
                    int extra) {

  int temp;
  int common_window = 0;
  int i = 0;
  BLOCKTYPE win1, win2;
  unsigned int maskType = 0;

  RESIDUAL_DATA_INTERN* rdi =
    (RESIDUAL_DATA_INTERN*) self->bsFrame[0].resData.residualDataIntern;
  
  RDI_CHANNEL* rdic_l = &(rdi->rdiChannel[channel]);
  RDI_CHANNEL* rdic_r = &(rdi->rdiChannel[channel+1]);
  Float *coef_buf[2];

  temp = s_GetBits(self->bitstream, 4); 
  common_window = s_GetBits(self->bitstream, 1); 

  if ( common_window != 1 )
    myexit("decodeCpe(): common_window must be 1"); 

  if ( common_window )
    {
      s_get_ics_info(&(rdic_l->win),
                     &(rdic_l->wndShape.this_bk),
                     rdic_l->group,
                     &(rdic_l->max_sfb),
                     &(rdic_l->info),
                     &(rdi->winmap[0]));

      
      maskType = getmask(&(rdic_l->info), rdic_l->group, rdic_l->max_sfb, rdic_l->mask); 

      
      memcpy(rdic_r, rdic_l, sizeof(RDI_CHANNEL));

    }
 
  s_huffdecode(ID_SCE,
               common_window,
               &(rdic_l->win),
               &(rdic_l->wndShape),
               rdic_l->cb_map,
               rdic_l->factors,
               rdic_l->group,
               &(rdic_l->hasmask),
               rdic_l->mask,
               &(rdic_l->max_sfb),
               &(rdic_l->tns),
               rdic_l->coef,
               &(rdic_l->info),
               &(rdi->winmap[0])
               );
  
  s_huffdecode(ID_SCE,
               common_window,
               &(rdic_r->win),
               &(rdic_r->wndShape),
               rdic_r->cb_map,
               rdic_r->factors,
               rdic_r->group,
               &(rdic_r->hasmask),
               rdic_r->mask,
               &(rdic_r->max_sfb),
               &(rdic_r->tns),
               rdic_r->coef,
               &(rdic_r->info),
               &(rdi->winmap[0])
               );


  coef_buf[0] = rdic_l->coef;
  coef_buf[1] = rdic_r->coef;

  
  synt( &(rdic_l->info), rdic_l->group, rdic_l->mask, coef_buf[1], coef_buf[0] );


  applyTools(rdic_l->coef,
             &rdic_l->tns,
             rdic_l->max_sfb,
             &rdic_l->info);

  applyTools(rdic_r->coef,
             &rdic_r->tns,
             rdic_r->max_sfb,
             &rdic_r->info);

  
  self->resMdctPresent[channel][channelFrame] = 1;

  if (extra==1) 
    assert(self->resBlocktype[channel][channelFrame] == rdic_l->win);
  else
    {
      self->resBlocktype[channel][channelFrame] = rdic_l->win;
      self->resBlocktype[channel + 1][channelFrame] = rdic_l->win;
    }

 
  for (i=0; i< BLOCK_LEN_LONG; i++) {
    self->resMDCT[channel][channelFrame][extra*MAX_MDCT_COEFFS + i]=
      coef_buf[0][i]; 
  }
  for (i=0; i< BLOCK_LEN_LONG; i++) { 
    self->resMDCT[channel + 1][channelFrame][extra*MAX_MDCT_COEFFS + i]=
      coef_buf[1][i]; 
  }

 return rdic_l->win;

}

void initResidualCore(spatialDec* self) {

  RESIDUAL_DATA_INTERN** rdi =
    (RESIDUAL_DATA_INTERN**) &(self->bsFrame[0].resData.residualDataIntern);

  *rdi = (RESIDUAL_DATA_INTERN*) calloc(1, sizeof(RESIDUAL_DATA_INTERN));


  if (self->residualCoding != 0) {
    s_huffbookinit(self->bsConfig.bsResidualSamplingFreqIndex);
  } else {
    if (self->arbitraryDownmix == 2) {
      s_huffbookinit(self->bsConfig.bsArbitraryDownmixResidualSamplingFreqIndex);
    }
  }

  (*rdi)->winmap[0] = s_win_seq_info[ONLYLONG_WINDOW];
  (*rdi)->winmap[1] = s_win_seq_info[ONLYLONG_WINDOW];
  (*rdi)->winmap[2] = s_win_seq_info[EIGHTSHORT_WINDOW];
  (*rdi)->winmap[3] = s_win_seq_info[ONLYLONG_WINDOW];


  return;
}
